"""First Stage Phonotactic Learner for Klamath data with the full set of suffixes, -a -tk and -Ca.
This learner includes MaxSTEM and IdhiSTEM to drive correct hiatus resolution.
Run this to get the initial weights to use with allsuffixesSecondStage.py
Charlie O'Hara 2/5/17
"""

import numpy as np
from math import exp, log
import scipy.optimize
import json
import random

#Import data from text file.
with open ("fullset.txt", "r") as myfile:
    data=myfile.read().splitlines()

#Fill inpout (input-output) with forms from data
#inpout is a dictionary that associates ordered pairs of morpheme tags with their surface forms. 
# (Stem-tag, Suffix-tag) : Surface-Form
output=[]
for x in data:
	output.append(x.split('\t')[0])
	
	
#Set of vowels considered with +high
HIGH=["i","I"]
#Set of vowels with -high
LOW=["a","e"]
#Set of vowels overall
VOWELS="a e i u I".split()
#Set of vowels, including long vowels
VOWLONGS="a e i u a: e: i: u:".split()
#List of the names of the constraints used in this simulation
CONSTRAINTS="Idhi maxv *e maxvin IdATR depv *I *Ctk IdHiin IdATRin *VV MaxSTEM IdhiSTEM".split()
#CONSTRAINTS="Idhi maxv *e maxvin IdATR depv *I *Ctk IdHiin IdATRin".split()
#Initial weights of constraints (here after phonotactic learning)
#Whether the constraints are faithfulness (f) or markedness (m), ties to whether they are biased high or low.  
CONTYPE="f f m f f f m m f f m f f".split()
#CONTYPE="f f m f f f m m f f".split()

#numcon represents the number of constraints in the grammar.
numcon=len(CONTYPE)
#SIGMAM is a constant that represents the plasticity of markedness constraints, the higher SIGMAM, the more markedness constraints can move.
SIGMAM=10
#SIGMAF does the same thing for faithfulness constraints.
SIGMAF=25
#mbias is the constraint weight that markedness constraints are biased towards---meant to be high, in order to cause restrictiveness
mbias=100
#sigma is used as a scaling factor for the prior which prevents frequencies of URs from getting too high.
sigma=.001
#Bound used to prevent constraint weights and UR frequencies from getting below too negative.
negative=-0.00000001



#Take in input (ordered pair of stem and suffix URs) , output (ordered pair of surface form and Correspondence relation) and WEIGHTS to find harmony score.
def harmony(input, outputpair, WEIGHTS):
	harmonyscore=0
	#find the violation vector of this input-output candidate
	viols=vios(input[0],input[1], outputpair)
	#Perform the dot product of the constraint weight and violation vectors.
	for x in range(len(viols)):
		harmonyscore+=WEIGHTS[x]*(-viols[x])
	#return harmony score
	return harmonyscore


#Create candidates that raise /e/ and /I/ to [i]. Take in an input string, FFC (a fully faithful correspondence relation), and a list of the indices of /e/ and /I/ in the input
def rise(input,FFC,lows):
	#Find the first index of a low vowel. 
	first=lows[0]
	#If there is just one index in lows, return the fully faithful candidate as well as the candidate that replaces the vowel at first with [i]
	if len(lows)==1:
		high="i"
		return [(input,FFC), ("%s%s%s"%(input[:first],high,input[first+1:]),FFC)]
	#If there are multiple indices in lows, this function runs recursively.
	else:
		high="i"
		#Return the union of the candidates that do not raise the vowel at the index first, and those that do; but remove the index from first.
		return rise(input,FFC,lows[1:])+ rise("%s%s%s"%(input[:first],high,input[first+1:]),FFC,lows[1:])
#Create candidates that lower /i/ and /I/ to [e]. Take in an input string, FFC (a fully faithful correspondence relation), and a list of the indices of a high vowel in the input	
def height(input,FFC, highs):
	#Find the first index of a high vowel. 
	first=highs[0]
	#If only one index is in high, return the faithful candidate and the candidate with lowering.
	if len(highs)==1:
		if input[first]=="i" or input[first]=="I":
			low="e"
		else:
			low=input[first]
		return [(input,FFC), ("%s%s%s"%(input[:first],low,input[first+1:]),FFC)]
	#If there are multiple indices in highs, this function runs recursively.
	else:
		if input[first]=="i" or input[first]=="I":
			low="e"
		else:
			low=input[first]
		return height(input,FFC,highs[1:])+ height("%s%s%s"%(input[:first],low,input[first+1:]),FFC,highs[1:])
#Given the prefix and suffix strings, as well as the outpair (the surface form and correspondence relation).	
def vios(stem,suf, outpair):
	input="%s%s"%(stem,suf)
	output=outpair[0]
	corresp=outpair[1]
	#change the correspondence pairs from the correspondence relation, and turn them into a dictionary.
	crelation={}
	for x in corresp:
		crelation[x[0]]=x[1]

	
	viols=[]
	#Id[hi] add a violation if a vowel's input correspondent does not match its output correspondent in height
	Dhigh=0
	for x in range(len(input)):
		if x in crelation:
			#Add a violation if input segment is high and its output correspondent is not.
			if input[x] in HIGH and output[crelation[x]] not in HIGH:
				Dhigh+=1
				
	for x in range(len(input)):
		if x in crelation:
			#Add a violation if input segment is not high and its output correspondent is high.
			if input[x] not in HIGH and output[crelation[x]] in HIGH:
				Dhigh+=1

	#append the number of violations of Id[hi] to the violation vector.
	if Dhigh>=0:
		viols.append(Dhigh)
	else:
		viols.append(0)
	

	
	#Max-V add a violation if a vowel is not linked to an output correspondent.
	MinV=0
	
	for x in range(len(input)):#add one for each vowel with no correspondence to any output segments.
		if x not in crelation:
			MinV+=1
	#append the number of violations of Max-V to the violation vector.
	if MinV>=0:
		viols.append(MinV)
	else:
		viols.append(0)
		
	#*e Assign a violation for each output segment with [e]. 
	e=0
	for x in range(len(output)):
		if output[x]=="e":
			#if x==len(output)-1:
			#	e+=1
			#elif output[x+1]!=":":
			e+=1
			
	#append the number of violations of *e to the violation vector.
	viols.append(e)

	#maxv/in Assign a violation if the initial vowel is not linked to an output correspondent.
	MinV=0
	firstvow=-1
	outsyl=filter(lambda x: x in VOWELS+[':'], output)
	for x in range(len(input)):
		if input[x] in VOWELS: #Find the first segment in the input that is a vowel, save this index as firstvow.
			if firstvow==-1:
				firstvow=x
				#If the first vowel in the input has no output correspondent, assign one violation mark.
				if x not in crelation:
					MinV+=1
	#append the number of violations of maxv/in to the violation vector.
	if MinV>=0:
		viols.append(MinV)
	else:
		viols.append(0)
	
	#IDATR Assign a violation if a vowel the ATR value of the input is not the same as its output correspondent.
	DATR=0
	
	for x in range(len(input)):
		if x in crelation:
		#Add a violation if input segment is -ATR and its output correspondent is not.
			if input[x]=="I" and output[crelation[x]] !="I":
				DATR+=1
				
	for x in range(len(input)):
		if x in crelation:
		#Add a violation if input segment is +ATR and its output correspondent is not.
			if input[x]!="I" and output[crelation[x]]=="I":
				DATR+=1

	#append the number of violations of IdATR to the violation vector.	
	if DATR>=0:
		viols.append(DATR)
	else:
		viols.append(0)
	
	#DepV Assign a violation if an output vowel has no input correspondent
	PlsV=0
	for x in range(len(output)): #Iterate across the output segments and count a violation for each segment without an input correspondent
		if x not in crelation.itervalues():
			PlsV+=1
	
	#append the number of violations of DepV to the violation vector.
	if PlsV>=0:
		viols.append(PlsV)
	else:
		viols.append(0)
	
	
	
	#*I Assign a violation for each output segment with [I]. 
	I=0
	for x in range(len(output)):
		if output[x]=="I":
			I+=1
	#append the number of violations of *e to the violation vector.
	viols.append(I)
	
	#Ctk Assign a violation for each nonphonotactically illicit CCC sequence in the output
	ctk=0
	if 'tk' in output:
		if output[output.index('tk')-1] not in VOWELS+[":","n"]:
			ctk+=1
	if 'c3t' in output:
		ctk+=1
	if 'n5t' in output:
		ctk+=1
	if '*wl' in output:
		ctk+=1
	if 'mntk' in output:
		ctk+=1
	#append the number of violations of *Ctk to the violation vector.
	viols.append(ctk)
	#Id[hi]/in Assign a violation if the initial vowel of the input's high value  is not the same as its output correspondent.
	Dhigh=0
	x=firstvow
	if x in crelation:
		if input[x] in HIGH and output[crelation[x]] not in HIGH:
			Dhigh+=1
		if input[x] not in HIGH and output[crelation[x]] in HIGH:
			Dhigh+=1

	#append the number of violations of Id[hi]/in to the violation vector.
	if Dhigh>=0:
		viols.append(Dhigh)
	else:
		viols.append(0)
	#IdATR/in Assign a violation if the initial vowel of the input's  ATR value  is not the same as its output correspondent.
	Dhigh=0
	x=firstvow
	if x in crelation:
		if input[x] =="I" and output[crelation[x]]!="I":
			Dhigh+=1
		if input[x]=="I" and output[crelation[x]]!="I":
			Dhigh+=1

	#append the number of violations of ID[ATR]/Hi to the violation vector.
	if Dhigh>=0:
		viols.append(Dhigh)
	else:
		viols.append(0)
		
		
	#*VV Assign a violation mark for two adjacent vowels.
	vvs=0
	for x in range(len(output)-1):
		if output[x] in VOWELS and output[x+1] in VOWELS:
			vvs+=1
	#append the number of violations of *VV to the violation vector.
			
	if vvs>=0:
		viols.append(vvs)
	else:
		viols.append(0)
		
	#MaxSTEM Assign violations for segments in the stem in the input with no output correspondent
	MinV=0
	for x in range(len(stem)):
		if x not in crelation:
			MinV+=1
			
#	MinV=0
#	for x in range(len(input)):
#		if input[x] in 'i I e'.split():
#			if x not in crelation:
#				MinV+=1		
	#append the number of violations of Max/STEM to the violation vector.
	if MinV>=0:
		viols.append(MinV)
	else:
		viols.append(0)
		
	#Id[hi]/STEM add a violation if a vowel's input correspondent does not match its output correspondent in height
	Dhigh=0
	for x in range(len(stem)):
		if x in crelation:
			#Add a violation if input segment is high and its output correspondent is not.
			if input[x] in HIGH and output[crelation[x]] not in HIGH:
				Dhigh+=1
				
	for x in range(len(stem)):
		if x in crelation:
			#Add a violation if input segment is not high and its output correspondent is high.
			if input[x] not in HIGH and output[crelation[x]] in HIGH:
				Dhigh+=1

	#append the number of violations of Id[hi] to the violation vector.
	if Dhigh>=0:
		viols.append(Dhigh)
	else:
		viols.append(0)
	

	return viols	
#Create candidates that epenthesize [a] . Take in an input string, FFC (a fully faithful correspondence relation), and blanks: a list of the indices of consonants followed by [t] or [C].	
def epenv(input,FFC,blanks):
	first=blanks[-1]
	#create a new correspondence relation to shift output indices to make up for the epenthesized vowel.
	new1=[]
	for j in FFC:
		i=j[1]
		k=j[0]
		if i<first:
			new1.append((k,i))
		elif i>=first:
			new1.append((k,i+1))
	newFFC=tuple(new1)
	#If there is just one index in blanks, return the fully faithful candidate as well as the candidate that inserts a vowel at first

	if len(blanks)==1:
		return [(input,FFC), ("%sa%s"%(input[:first],input[first:]),newFFC)]
	#If there are multiple indices in blanks, this function runs recursively.
	else:
		return epenv(input,FFC,blanks[:-1])+ epenv("%sa%s"%(input[:first],input[first:]),newFFC,blanks[:-1])

#Create candidates that delete [a] . Take in an input string, FFC (a fully faithful correspondence relation), and vowels: a list of the indices of all the vowels in the input.	
def delv(input,FFC, vowels):
	#If there are no indices in vowels, return just the faithful candidate.
	if len(vowels)==0:
		return [(input,FFC)]
	
	
	else:
		first=vowels[-1]
		#Develop a new correspondence relation shifted for the deletion of vowels
		new1=[]
		for j in FFC:
			i=j[1]
			k=j[0]
			if i<first:
				new1.append((k,i))
			elif i!=first:
				new1.append((k,i-1))
		new=tuple(new1)
		#Create deletion candidates if the vowels are marked (here /e/ and /I/)
		if input[first]=='e' or input[first]=='I':
			#If there is just one index in vowels, return the fully faithful candidate as well as the candidate that deletes the vowel at first
			if len(vowels)==1:
				return [(input,FFC), ("%s%s"%(input[:first],input[first+1:]),new)]
			#If there are multiple indices in vowels, this function runs recursively.
			else:
				return delv(input,FFC,vowels[:-1])+ delv("%s%s"%(input[:first],input[first+1:]),new,vowels[:-1])
		#Create deletion candidates for vowels in VV sequences.
		#If one of the vowels is initial in the input.
		elif first==0:
			if  input[first+1] in VOWELS: 
				#If there is just one index in vowels, return the fully faithful candidate as well as the candidate that deletes the vowel at first
				if len(vowels)==1:
					return [(input,FFC), ("%s%s"%(input[:first],input[first+1:]),new)]
				else:#If there are multiple indices in vowels, this function runs recursively.
					return delv(input,FFC,vowels[:-1])+ delv('%s%s'%(input[:first],input[first+1:]),new,vowels[:-1])
			else:
				return delv(input,FFC,vowels[:-1])
		#If one of the vowels i final.
		elif first==len(input)-1:
			if input[first-1] in VOWELS: 
				if len(vowels)==1:#If there are multiple indices in vowels, this function runs recursively.
					return [(input,FFC), ("%s%s"%(input[:first],input[first+1:]),new)]
				else:#If there are multiple indices in vowels, this function runs recursively.
					return delv(input,FFC,vowels[:-1])+ delv("%s%s"%(input[:first],input[first+1:]),new,vowels[:-1])
			else:
				return delv(input,FFC,vowels[:-1])
		#If the vowels are somewhere in he middle.
		elif input[first-1]  in VOWELS or input[first+1] in VOWELS: 
			if len(vowels)==1:
				return [(input,FFC), ('%s%s'%(input[:first],input[first+1:]),new)]
			else:
				return delv(input,FFC,vowels[:-1])+ delv('%s%s'%(input[:first],input[first+1:]),new,vowels[:-1])
			
		else:
			return [(input,FFC)]
			
#Take in an input string and a correspondence relation FFC. 		
def cands2(input,FFC):
	candset=[]
	#Create lowering of vowels
	highs=[]
	#Find the indices of each high vowel.
	for x in range(len(input)):
		if input[x] in HIGH:
			highs.append(x)
	if len(highs)==0:
		candset.append((input,FFC))
	else:# If there are high vowels call the height function which creates all permutations of lowering of high vowels.
		candset=candset+height(input,FFC,highs)
	#print candset
	#Create violations of Maxv 
	vs=[]
	#Find the indices of each vowel in the input word.
	for x in range(len(input)):
		if input[x] in VOWELS:
			vs.append(x)
	#Find the set of all permutations of markedness driven deletion 
	candset=candset+delv(input,FFC,vs)

	#Create rising of vowels
	lows=[]
	#Find indices of e and I
	for x in range(len(input)):
		if input[x] in ["e","I"]:
			lows.append(x)
	if len(lows)==0:
		candset.append((input,FFC))
	else: # If there are any vowels of this sort, return permutations of rising across the e and I vowels.
		candset=candset+rise(input,FFC,lows)
	# Create violations of DepV
	blanks=[]
	#Gather indices of consonants before t and C: blanks
	for x in range(len(input)):
		if x!=0:
			if input[x-1] not in VOWELS and input[x]=='t':
				blanks.append(x)
			if input[x-1] not in VOWELS and input[x]=='C':
				blanks.append(x)			
	if len(blanks)!=0: # Find all permutations of markedness driven epenthesis.
		candset=candset+epenv(input,FFC,blanks)
	
	return list(set(candset))	

canddict={}
#Given an input pair (stem, suffix), return a set of potential candidates. 
def cands(inpair):
	input="%s%s"%(inpair[0],inpair[1])
	sets=[]
	#Check if the set of candidates has been found already and is saved in canddict, so we don't have to run this too many times.
	if input in canddict:
		return canddict[input]
	else:
		#Create a correspondence relation with input and output.
		FFC=()
		for x in range(len(input)):
			FFC=FFC+((x,x),)
		#Call the cands2 function to generate a variety of output candidates.
		candset=cands2(input,FFC)
		# For each candidate in candset, add the set of candidates generated by cands2 for this candidate.
		for x in candset:
			sets=list(set(sets+cands2(x[0],x[1])))
		#Save the set of candidates in the canddict.
		canddict[input]=sets
		return sets
		
#Given inputpair and the WEIGHTS of constraints and find the probability of each candidate given the input stem and suffix URs.
def probs(inputpair,WEIGHTS):
	candset=cands(inputpair)
	htotal=0
	harmdict={}
	#For each candidate find the harmony score and add up the total harmony score.
	for x in list(set(candset)):
		harmdict[x]=exp(harmony(inputpair,x,WEIGHTS))
		htotal+=harmdict[x]
	for x in harmdict: #Calculate the probability for each candidate
		if htotal<1e-150: #If all forms have very low probability give all 0 probability
			harmdict[x]=0
		else:
			harmdict[x]=harmdict[x]/htotal 
	return harmdict
	
#Calculate the L2 gaussian prior, the sum of the difference of each weight from it's bias point (mbias or 0) squared divided by the associated SIGMA value
def prior(WEIGHTS):
	total=0
	for i in range(len(WEIGHTS)):
		w=WEIGHTS[i]
		if CONTYPE[i]=='f':
			total+=w**2/(2*SIGMAF**2)
		else:
			total+=(w-mbias)**2/(2*SIGMAM**2)
	return total	
	
#Calculate the sum squared of all frequencies of underlying representations.	
def priorl(lexprob):
	sum=0
	for x in lexprob:
		sum+=x**2
	return sum
	
#Calculate the objective function for the given grammar.
def objective(parameters):
	#Separate the constraint weights (w) from the UR frequencies (lvect)
	w=parameters[:numcon]
	probsum=0
	# For each piece of data find the stem and suffix morpheme tags
	for surfaceform in output:
		
		outputprobs={}
		probcand=probs((surfaceform,""),w)
		for x in probcand:
			if x[0] in outputprobs:
				outputprobs[x[0]]+=probcand[x]
			else:
				outputprobs[x[0]]=probcand[x]
		likelihood=outputprobs[surfaceform]
			
	
		#IF likelihood of the winner is super low, print error and mess up the objective
		if likelihood<=1e-150:

			probsum+=float('inf')
		else:	
			#add the negative log likelihood to probsum.
			probsum+=-log(likelihood)
	#As a failsafe, if some parameter is outside of its bounds, make the objective really big to let the algorithm know it messed up
	for x in parameters:
		if x<negative:
			return float("inf")
	#print prior(w)
	#Return the sum of negative log likelihood with the prior on constraints and morpheme frequencies.
	return probsum+prior(w)
#Calculate the gradient of the objective function for each dimension
def gradient(parameters):
	grad=[]
	WEIGHTS=parameters
	#START BY FINDING GRADIENTS FOR EACH CONSTRAINT WEIGHT
	
	O=[0]*len(WEIGHTS)
	E=[0]*len(WEIGHTS)
	
	for y in output:
		FFC=()
		for x in range(len(y)):
			FFC=FFC+((x,x),)
		v=vios(y,'', (y,FFC))
		
		for x in range(len(v)):
			O[x]+=v[x]
	for y in output:
		probdist=probs((y,''),WEIGHTS)
		for cand in probdist:
			v=vios(y,'',cand)
			if v[0]!=v[9]:
			for x in range(len(v)):
				E[x]+=probdist[cand]*v[x]

	for x in range(len(O)):
		if CONTYPE[x]=='f':
			grad.append((O[x]-E[x]+WEIGHTS[x]/(SIGMAF**2)))
		else:
			grad.append((O[x]-E[x]+(WEIGHTS[x]-mbias)/(SIGMAM**2)))
	
	print "---CONSTRAINTS------"
	printconstraints(parameters[:numcon])
	
	
	print " gradient	%s"%np.linalg.norm(grad)
	print " objective %s"%objective(parameters)
	print grad
	return np.asarray(grad)

#Given a dictionary d print those forms that have received less than .9 probability.
def printliky(d):
	y=0
	string=""
	if any(v < .9 for v in d.itervalues()):
		print "---SURFACE FORMS WITH LESS THAN .9 PROBABILITY----"	
	for x in d:
		if d[x]<.9:
			if len(x)<8:
				string+="	%s		%.2f,"%(x, d[x])
			else:
				string+="	%s	%.2f,"%(x, d[x])
			if y%3==2:
				print "%s"%(string)
				string=""
			y+=1
	
	print "%s" %string
	
	
#Print the probability of each output form given the grammar.	
def printprob(parameter):
	ew1=parameter[:numcon]
	lprevect=paramToVect(parameter[numcon:],lexvectkey,paramlex[1])
	URfreqs={}
	for i in range(len(lexvectkey)):
		lam,morph=lexvectkey[i]
		val=lprevect[i]
		if lam in URfreqs:
			URfreqs[lam][morph]=val
		else:
			URfreqs[lam]={morph:val}
	lexyprobs=pLEX(URfreqs)
	likylist={}
	for y in inpout:
		prob={}
		#print y
		lpre=y[0]
		lsuf=y[1]
		#print LEX
		for pre in LEX[lpre]:
			for suf in LEX[lsuf]:
				pbr=lexyprobs[lpre][pre]*lexyprobs[lsuf][suf]
				input="%s%s"%(pre,suf)
				proby=probs((pre,suf),ew1)
				for yum in proby:
					x=yum[0]
					if x in prob:
						prob[x]+=pbr*proby[yum]
					else:
						prob[x]=pbr*proby[yum]
		likylist[inpout[y]]=prob[inpout[y]]
	printliky(likylist)
#Print the weight of each constraint, with its name.
def printconstraints(d):
	string=""
	for x in range(numcon): 
		if len(CONSTRAINTS[x])<8:
			string+="	%s		%.2f,"%(CONSTRAINTS[x], d[x])
		else:
			string+="	%s	%.2f,"%(CONSTRAINTS[x], d[x])
		if x%4==3:
			print "%s"%(string)
			string=""
	print "%s"%(string)
#Given a dictionary of probabilities of URs for each morpheme, print them.
def printdict(d):
	for x in d:
		if len(d[x])>1:
			string=""
			for y in d[x]:
				if len(y)<8:
					string+="	%s		%.2f,"%(y, d[x][y])
				else:
					string+="	%s	%.2f,"%(y, d[x][y])
			print "%s"%(string)


def vectToParam(vect,key):
	param=[]
	paramkey=[]
	sum=0
	URfreqs={}
	for i in range(len(lexvectkey)):
		lam,morph=lexvectkey[i]
		val=vect[i]
		if lam in URfreqs:
			URfreqs[lam][morph]=val
		else:
			URfreqs[lam]={morph:val}
	for x in URfreqs:
		if len(URfreqs[x])>1:
			for y in URfreqs[x]:
				param.append(URfreqs[x][y])
				paramkey.append((x,y))

	return[param,paramkey]
#Given a dictionary of UR frequencies, return a vector.
def paramToVect(param,vkey,pkey):
	vect=[]
	
	for y in vkey:
		x=(y[0],y[1])
		if x in pkey:
			value=param[pkey.index(x)]
			vect.append(value)
			
		else:
			vect.append(1.0)
			
	return vect

test=[]
for w in range(len(CONTYPE)):
	if CONTYPE[w]=='f':
		test.append(0)
	else:
		test.append(mbias)
WEIGHTS=test
#Create bounds for the optimize algorithm (don't let constraints or frequencies get negative or weights get above 200)
wbound=(0,200)
bounds=[wbound,]*(numcon)

bigbound=bounds#+lexbond





#Run the SLSQP optimize algorithm on the objective function using the gradient function as a jacobian.
res=scipy.optimize.minimize(objective,np.asarray(test), method='SLSQP',bounds=bigbound,jac=gradient,  options={'maxiter':1000,'ftol':1e-50, 'eps':1e-5, 'disp': True})

w1=res.x
print w1

ew1=[]
for x in w1[:numcon]:
	ew1.append(exp(x))
ew1=w1[:numcon]
printconstraints(ew1)
